#!/bin/bash

ffoo import exits

#
# Name of pretty-printer module
#
# Friendly name of module used for decorating output.  For
# example, if the value is NAME, a _pretty_NAME module must
# exist and be importable.  Otherwise pretty will fall back
# to plain, which is also the default value.
#
FFOO_PRETTY=${FFOO_PRETTY:-plain}


__ffoo_pretty__init() {
    #
    # Import proper submodule
    #
    if ffoo try_import _pretty_${FFOO_PRETTY};
    then
        ffoo import _pretty_${FFOO_PRETTY}
        return 0
    else
        warn "falling back to _pretty_plain"
        FFOO_PRETTY=plain
        ffoo import _pretty_${FFOO_PRETTY}
    fi
}


##--------------------##
## PRINTING FRONT-END ##
##--------------------##

debug() {
    #
    # You already know what it does
    #
    # BTW, following are equivalent:
    #
    #     debug "var1=$var1" "var2=$var2" "result=$result"
    #     debug -v var1 var2 result
    #     debug -v "var@ result"
    #
    $FFOO_DEBUG || return 0
    __echo "$@"
}


debug_pipe() {
    #
    # Debug the whole pipe.
    #
    grep . \
        | while IFS= read line;
          do
              debug "|$1: '$line'"
              echos "$line"
          done
}


die() {
    #
    # A fatal error
    #
    __echo -t
    __echo "$@"
    exits_error
}


echos() {
    #
    # Safer version of echo able to echo "-n"
    #
    # Traditional echo is broken in that it does not
    # distinguish between string to print and its own switches
    # (-e, -E or -n), leading to unexpected behavior.
    #
    # This echo version circumvents this by using printf.
    #
    printf -- '%s\n' "$@"
}


usage_is() {
    #
    # Echo out usage patterns and (by default) `exit 2`
    #
    #     usage_is [-e STATUS] [-E] pattern [patern...]
    #
    # Each pattern is prefixed by "usage: " and a resolved
    # script/function name to produce traditional usage hint.
    # By default, will exit with status *FFOO_EXITS_USAGE*, you
    # can use `-e STATUS` to specify other number. `-k` can be
    # used as shorthand for `-e 0` (e.g. if user asked for
    # the patterns)
    #
    # Use `-E` to prevent exiting; in that case, the status
    # that would be fed to `exit` will be used as exit status
    # of this function.
    #
    # Use "--" to delimit end of arguments processed by usage_is.
    #
    # Recommended usage is to define usage() in your script and
    # use this in its body.  That way you only need to define
    # usage patterns once and skip to them from any place where
    # you detect incorrect usage.
    #
    local es=$FFOO_EXITS_USAGE
    local doexit=true
    while true; do case "$1" in
        -e) es="$2";            shift 2      ;;
        -E) doexit=false;       shift        ;;
        -k) es=$FFOO_EXITS_OK;  shift        ;;
        --)                     shift; break ;;
        *)                      break        ;;
    esac done
    __echo -u "$@";
    $doexit && exits "$es"
    return "$es"
}


think() {
    #
    # If verbose is on, think loud
    #
    # Use "-l" to split every parameter to separate line, (useful
    # or longer warnings)
    #
    $FFOO_VERBOSE || return 0
    __echo "$@"
}


warn() {
    #
    # Warn them
    #
    # Use "-l" to split every parameter to separate line, (useful
    # or longer warnings)
    #
    # Resist  temptation to prefix "warning" or similar BS here.
    # STDERR is *made* for warnings.
    #
    __echo "$@"
}


##----------##
## BACK-END ##
##----------##


__cat() {
    #
    # `cat` but without starting a process
    #
    # Used to avoid spanning new process where stream handler is chosen
    # based on some logic
    #
    while IFS= read line;
    do echos "$line"; done
}


__cut_stack() {
    #
    # call stack except the head that fits pattern $1
    #
    # E.g. if call stack is
    #
    #     foo1
    #     foo2
    #     bar
    #     bazz
    #     food
    #
    # call `__cut_stack ^foo` returns
    #
    #     bar
    #     bazz
    #     food
    #
    # i.e. will throw away all lines from start of stdin that
    # do match $1.  First line that does not match is printed,
    # as well as any other lines (no matter the match).
    #
    # This is used to hide boring part of stack that is present
    # as result of calls between our internal functions.
    #
    local patt=${1:-"^__"}
    echos "${FUNCNAME[@]}" \
      | perl -ne "
            my \$checking = 1;
            while (<>) {
                next if (\$checking and m/$patt/);
                \$checking = 0;
                print qq/\$_\\n/;
            }
        "
}


__echo() {
    #
    # A smarter echo backend
    #
    # A smarter backend for debug, warn, think, die and
    # usage_is.
    #
    #   -c cmd      echo output of a command
    #   -f file     echo output of a file (- for stdin)
    #   -l line [line...] echo each line separately
    #   -t          add stack trace to output
    #   -u patt [patt...] convert each patt to usage pattern
    #   -v var [var...] show contents of each var
    #
    local frontend  # who (of pretty.sh) was called (=> prettyprinter choice)
    local caller    # which user's function (or script) called it
    frontend="$(__cut_stack '^__echo' | head -1 )"
    caller="$(
        __cut_stack \
            '^__echo|^warn$|^debug_pipe$|^debug$|^die$|^usage$|^usage_is$|^think$' \
            | head -1
    )"
    test "$frontend" == "main" && frontend=$(basename $0)
    test "$caller" == "main" && caller=$(basename $0)
    test "$frontend" == "usage" && frontend=usage_is
    local provider       # which provider (__echo_*()) to use
    while true; do case $1 in
        -c|--cmd)   provider=cmd;   shift; break ;;
        -f|--files) provider=files; shift; break ;;
        -l|--lines) provider=lines; shift; break ;;
        -t|--trace) provider=trace; shift; break ;;
        -u|--usage) provider=usage; shift; break ;;
        -v|--vars)  provider=vars;  shift; break ;;
        *)          provider=args;         break ;;
    esac done
    case $frontend in
        debug)    ;&
        die)      ;&
        think)    ;&
        usage_is) ;&
        warn)
            __echo_$provider "$@" | _pretty_$frontend >&2
            ;;
        *)
            echo "do not call __echo* directly: $frontend" >&2
            exits_usage
            ;;
    esac
}


__echo_args() {
    #
    # The simplest (but safe) printing of args
    #
    echos "$*"
}


__echo_cmd() {
    #
    # Print command line, launch it and report exit status
    #
    local c=\$
    test $(id -u) -eq 0 && c=\#
    echo "$c $*"
    $@
    echo "^ exit status: $?"
}


__echo_files() {
    #
    # Print names and contents of existing files
    #
    local fp
    for fp in "$@";
    do
        test -s "$fp" || continue
        echo "-- $fp --"
        cat $fp
    done
}


__echo_lines() {
    #
    # Echo each argument as a separate line
    #
    local l;
    for l in "$@"; do __echo "$l"; done
}


__echo_trace() {
    #
    # Print "decorated" call trace (only in debug mode)
    #
    $FFOO_DEBUG || return 0
    echo "== trace =="
    caller | sed -e 's/^/-> /g'
}


__echo_usage() {
    #
    # Print each argument as usage pattern
    #
    local u
    for u in "$@";
    do echo "usage: $caller $u"
    done
}


__echo_vars() {
    #
    # Report value of each named variable
    #
    # If the last char is '@', try to use it as wildcard
    #
    local varname
    for varname in "$@";
    do
        if ! __is_word "$varname";
        then
            warn "unsafe value skipped: $varname";
            continue
        fi
        __echo "$varname='$(echos "${!varname}")'"
    done
}


__is_word() {
    #
    # Check if $1 contains only alphanumeric chars or _
    #
    local tainted="$1"
    local clean=$(tr -c -d '_[:alnum:]' <<< "$tainted")
    test "$tainted" = "$clean"
}
